from clickhouse_connect import driver_name
from AICORELibrary.clickhouse.cc_sqlalchemy.datatypes.base import schema_types

# pylint: disable=invalid-name
dialect_name = driver_name
ischema_names = schema_types
